
/* COBOL Preprocessor, constructs to deal with (+ marks done):
   + margins
   - COPY
   - REPLACE
   - other preprocessing statements
   + comments (* and / in column 7)
   - line continuations
   + other indicators
   - case sensitivity
   + comment entries
   - EXEC
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#define MAX_LINE_WIDTH 1000

enum divisions {NONE, IDENTIFICATION, ENVIRONMENT, DATA, PROCEDURE};

static char output_line[MAX_LINE_WIDTH+1];
static char buf[MAX_LINE_WIDTH+1];
static int inside_comment_entry;
static int line;
static int division;
static int nr_newlines;
static int quote72;

/* Utility functions. */
/*
static void strtolower(char * p)
{
    int i;

    for (i = 0; i < strlen(p); i++) {
	p[i] = tolower(p[i]);
    }
}
*/

static void strtoupper(char * p)
{
    int i;

    for (i = 0; i < strlen(p); i++) {
	p[i] = toupper(p[i]);
    }
}

static void clear_line(void)
{
    int i;
    for (i = 0; i < 72; i++) {
        buf[i] = ' ';
    }
}

static int area_a_empty(void)
{
    int i;

    for (i = 7; i < 11; i++) {
	if (buf[i] != ' ') return 0;
    }
    return 1;
}

/* Strip the margins to ease parsing. */
static void strip_margins(void)
{
    int i;

    /* Change the left margin to whitespace. */
    for (i = 0; i < 6; i++) {
	buf[i] = ' ';
    }
    /* Remove the right margin altogether. */
    buf[72] = '\0';
}

/* Use inline comments: "*>". */
static int is_comment(void)
{
    return buf[0] == '*';
}

static void make_comment(void)
{
    buf[0] = '*';
    buf[1] = '>';
    buf[6] = ' ';
}

/* Make sure comments and multiplication/division don't conflict. */
static void handle_comments(void)
{
    if (buf[6] == '*' || buf[6] == '/') {
	make_comment();
    }
}

/* Determine wether the last nonnumeric literal is closed. */
static int literal_closed(void)
{
    int i, ch, inside;

    i = 0;
    inside = 0;
    while ((ch = output_line[i]) != '\0') {
	if (ch == '\"') {
	    if (inside) {
		if (output_line[i+1] == '\"') {
			i += 2;
		} else {
			inside = 0;
			i++;
		}
	    } else {
		inside = 1;
		i++;
	    }
	} else {
	    i++;
	}
    }
    return !inside;
}

/* I have a hard time comprehending the VS COBOL II definition for this.
   The definition in the manual is ambiguous, it is not clear wether a
   quotation mark column in 72 overides all other continuation line
   definitions or not. Also I cannot find this "72" definition in any
   other manual, I will pretend it does not exist and see what happens.
 */
static void handle_continuations(void)
{
    int i;
    char * p;

    if (buf[6] == '-') {
	if (!area_a_empty()) {
	    fprintf(stderr, "error: Line %d is not a legal continuation line.\n", line);
	    exit(1);
	}
	if (!literal_closed() || quote72) {
	    p = index(buf, '\"');
	    if (p == 0) {
		fprintf(stderr, "error: Line %d is not a legal nonnumeric literal continuation line.\n", line);
		exit(1);
	    }
	    strcat(output_line, p + 1);
	} else {
	    for (i = strlen(output_line) - 1; i >= 0; i--) {
		if (output_line[i] != ' ') {
		    output_line[i+1] =  '\0';
		    break;
		}
	    }
	    for (i = 7; i < strlen(buf); i++) {
		if (buf[i] != ' ') break;
	    }
	    strcat(output_line, buf + i);
	}
	nr_newlines++;
    }

    /* Read the manual to see what this is about. */
    quote72 = buf[70] != '\"' && buf[71] == '\"';
}

/* Skip things like Y and C until I find documentation on them. */
static void handle_other_indicators(void)
{
    if (!is_comment() && buf[6] != '-') {
	buf[6] = ' ';
    }
}

/* Set a flag when we encounter certain types of area a. */
static void handle_area_a_lines(void)
{
    int m, n;
    static char first_word[MAX_LINE_WIDTH + 1];

    if (is_comment()) return;
    if (area_a_empty()) return;

    m = strspn(buf, " ");
    n = strcspn(buf + m, ".");
    strncpy(first_word, buf + m, n);
    strtoupper(first_word);
    first_word[n] = '\0';

    if (strcmp(first_word, "IDENTIFICATION DIVISION") == 0 ||
        strcmp(first_word, "ID DIVISION") == 0) {
	division = IDENTIFICATION;
    } else if (strcmp(first_word, "ENVIRONMENT DIVISION") == 0) {
	division = ENVIRONMENT;
    } else if (strcmp(first_word, "DATA DIVISION") == 0) {
	division = DATA;
    } else if (strcmp(first_word, "PROCEDURE DIVISION") == 0) {
	division = PROCEDURE;
    }

    if (division == IDENTIFICATION) {
	if (strcmp(first_word, "AUTHOR") == 0 ||
            strcmp(first_word, "INSTALLATION") == 0 ||
            strcmp(first_word, "DATE-WRITTEN") == 0 ||
            strcmp(first_word, "DATE-COMPILED") == 0 ||
            strcmp(first_word, "SECURITY") == 0) {
	    clear_line();
	    inside_comment_entry = 1;
	}
    }
}

/* Get rid of comment entries, they are going to become obsolete anwyay. */
static void handle_comment_entries(void)
{
    if (is_comment()) return;

    if (inside_comment_entry) {
	if (area_a_empty()) {
	    clear_line();
	} else {
	    inside_comment_entry = 0;
	}
    }
}

static void flush_output_line(void)
{
    int i;
    fputs(output_line, stdout);
    for (i = 0; i < nr_newlines; i++) {
	printf("\n");
    }
    nr_newlines = 1;
}

/* Read a line from stdin, preproces, write to stdout and repeat till EOF. */
int main(int argc, char **argv)
{
    int len, i;

    line = 1;
    inside_comment_entry = 0;
    division = NONE;
    nr_newlines = 1;
    quote72 = 0;

    /* Make sure output_line always contains a valid string. */
    output_line[0] = '\0';

    while (fgets(buf, MAX_LINE_WIDTH, stdin) != 0) {
	len = strlen(buf);
	/* Make sure the line is always the right size. */
	if (len != 81) {
	    for (i = len; i < 80; i++) {
		buf[i] = ' ';
	    }
	    buf[80] = '\0';
	}
	strip_margins();
	handle_comments();
	handle_comment_entries();
	handle_area_a_lines();
	handle_continuations();
	handle_other_indicators();
	if (buf[6] != '-') {
	    if (line != 1) {
		flush_output_line();
	    }
	    strcpy(output_line, buf);
	}
	line++;
    }
    flush_output_line();

    return 0;
}

