
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttparse.h"

#include "VSCScan.h"
#include "VSCParse.h"

#define MAX_KW_SIZE 1000

static BOOL is_whitespace(int ch)
{
	return ch <= ' ' && ch != '\0';
}

TTterm TTparseVSC_LAYOUT(void)
{
	const char *buf = TTparse_get_string();
	int len;
	int old_len;
	
	len = 0;
	old_len = len - 1;

	while (len > old_len)
	{
		old_len = len;
		if (is_whitespace(buf[len])) {
			len++;
	        } else if ((buf[len] == ',' || buf[len] == ';') &&
                           is_whitespace(buf[len+1])) {
			len += 2;
		} else if (buf[len] == '*' && buf[len+1] == '>') {
			len += strcspn(buf + len, "\n");
		}
	}

	return TTparse_word((size_t)len);
}

static TTterm scan_word(size_t len)
{
	TTterm t1, t2;

	t1 = TTparse_word(len);
/*
fprintf(stderr, "Token: ");
TTdump_string(stderr, t1);
fprintf(stderr, "\n");
*/
	t2 = TTparseVSC_LAYOUT();
	return TTbuild_sandwich(TTbuild_empty_string(), t1, t2);
}

TTterm TTparseVSC_KEYWORD(const char *name)
{
	const char *buf;
	size_t len = 0;
	BOOL possible_identifier = FALSE;
	int ch, ch2;

	buf = TTparse_get_string();

	while ((ch = name[len]) != '\0') {
		ch2 = buf[len];
		if (ch2 == 0) {
			return 0;
		}
		if (tolower(ch) != tolower(ch2)) {
			return 0;
		}
		if (isalnum(ch)) {
			possible_identifier = TRUE;
		}
		len++;
	}
	ch = buf[len];
	if (possible_identifier && (isalnum(ch) || ch == '-' || ch == '_'))
		return 0;
	return scan_word(len);
}

TTterm
TTparseVSC_lex_picture_string( void )
{
/* <PIC>([^ \t\n,;.]|([,;.][^ \t\n]))+ */
    const char * buf;
    size_t len;
    
    buf = TTparse_get_string();

    len = strcspn(buf, " \t\n");
    if (len > 0)
    {
        if (buf[len-1]  == '.')
        {
            len--;
        }
        if (len > 0)
        {
            return TTbuildVSC_lex_picture_string(scan_word(len));
        }
    }
    return 0;
}

TTterm
TTparseVSC_lex_quoted_pseudo_text( void )
{
/* "=="[^=]*"==" */
    const char * buf;
    size_t len;
    
    buf = TTparse_get_string();

    len = 0;
    if (buf[len] == '=' && buf[len+1] == '=')
    {
        len += 2;
        len += strcspn(buf+len, "=");
        if (buf[len] == '=' && buf[len+1] == '=')
        {
            len += 2;
            return TTbuildVSC_lex_quoted_pseudo_text(scan_word(len));
        }
    }
    return 0;
}

TTterm
TTparseVSC_lex_nonnumeric( void )
{
/* ["][^"\n]*["] */
/* ['][^'\n]*['] */
    const char * buf;
    size_t len, old_len;

    buf = TTparse_get_string();

    len = 0;
    if (buf[len] == '\"')
    {
        len++;
	do {
		old_len = len;
        	len += strcspn(buf+len, "\"");
		if (buf[len] == '\"' && buf[len+1] == '\"') len += 2;
	} while (old_len != len);
        if (buf[len] == '\"')
        {
            len++;
            return TTbuildVSC_lex_nonnumeric(scan_word(len));
        }
    }
    else if (buf[len] == '\'')
    {
        len++;
	do {
		old_len = len;
        	len += strcspn(buf+len, "\'");
		if (buf[len] == '\'' && buf[len+1] == '\'') len += 2;
	} while (old_len != len);
        if (buf[len] == '\'')
        {
            len++;
            return TTbuildVSC_lex_nonnumeric(scan_word(len));
        }
    }
    return 0;
}

TTterm
TTparseVSC_lex_cobword( void )
{
/* [A-Za-z0-9\-]+ */
    const char * buf;
    size_t len, i;
    char tmp[MAX_KW_SIZE+1];

    buf = TTparse_get_string();

    len = 0;
    while (isalnum((int)buf[len]) || buf[len] == '-' || buf[len] == '_')
    {
        len++;
    }
    if (len > 0)
    {
        if (len < MAX_KW_SIZE)
        {
/*            strncpy(tmp, buf, len);*/
	    for (i = 0; i < len; i++) {
		tmp[i] = toupper(buf[i]);
	    }
            tmp[len] = '\0';
            if (VSCisKeyword(tmp)) return 0;
        }
        return TTbuildVSC_lex_cobword(scan_word(len));
    }
    return 0;
}

TTterm
TTparseVSC_lex_num( void )
{
/* [\+\-]?(([0-9]*[\.\,][0-9]+)|[0-9]+) */

    const char * buf;
    size_t len;
    BOOL hasdigit = FALSE;
    int ch;

    buf = TTparse_get_string();

    len = 0;
    ch = buf[len];
    if (ch == '-' || ch == '+')
    {
        len++;
    }
    while (isdigit((int)buf[len]))
    {
        len++;
        hasdigit = TRUE;
    }
    ch = buf[len];
    if (ch == '.' || ch == ',')
    {
        len++;
        if (!isdigit((int)buf[len]))
        {
            len--;
        }
        else
        {
            while (isdigit((int)buf[len]))
            {
                len++;
                hasdigit = TRUE;
            }
        }
    }
    if (hasdigit && !isalnum((int)buf[len]) && buf[len] != '-')
    {
        assert(len > 0);
        return TTbuildVSC_lex_num(scan_word(len));
    }
    return 0;
}

TTterm TTparseVSC_DISJUNCTION(TTfun fun, size_t n, TTterm(**parsers) (void))
{
	TTterm result;

	if (fun == TTfunVSC_fd_entry_alt_characters_records &&
	    TTparse_guard(TTparseVSC_fd_entry_record_clause)) {
		result = 0;
	} else {
		result = TTparse_disjunction(fun, n, parsers);
	}
/*
fprintf(stderr, "%d %s\n", result != 0, fun->name);
*/
	return result;
}

TTterm TTparseVSC_CONJUNCTION(TTfun fun, TTterm(**parsers) (void))
{
	TTterm result;
	result = TTparse_conjunction(fun, parsers);
/*
fprintf(stderr, "%d %s\n", result != 0, fun->name);
*/
	return result;
}

